//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// デバイスの基本クラスなど
//

#include "device.h"
#include "mpu.h"
#include "scheduler.h"

//
// Device
//

// コンストラクタ
Device::Device(uint objid_)
	: inherited(objid_)
{
}

// デストラクタ
Device::~Device()
{
}

bool
Device::Create()
{
	return true;
}

bool
Device::Init()
{
	mpu = GetMPUDevice();
	scheduler = GetScheduler();

	return true;
}

bool
Device::Apply()
{
	return true;
}

void
Device::ResetHard(bool poweron)
{
}

void
Device::PowerOff()
{
}

// ログ表示。
// VM デバイスでは、仮想時間、PC、オブジェクト名を表示する。
void
Device::putlogn(const char *fmt, ...) const
{
	char buf[1024];
	va_list ap;
	uint64 vt;
	int len;

	vt = scheduler->GetVirtTime();
	len = snprintf(buf, sizeof(buf), "%4u.%03u'%03u'%03u %08x %s ",
		(uint)(vt / 1000 / 1000 / 1000),
		(uint)((vt / 1000 / 1000) % 1000),
		(uint)((vt / 1000) % 1000),
		(uint)(vt % 1000),
		mpu->GetPPC(),
		GetName().c_str());

	va_start(ap, fmt);
	vsnprintf(buf + len, sizeof(buf) - len, fmt, ap);
	va_end(ap);

	WriteLog(buf);
}


//
// IODevice
//

// コンストラクタ
IODevice::IODevice(uint objid_)
	: inherited(objid_)
{
}

// デストラクタ
IODevice::~IODevice()
{
}

// 初期化
bool
IODevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	return true;
}

busdata
IODevice::Read8(uint32 addr)
{
	return 0xff;
}

busdata
IODevice::Read16(uint32 addr)
{
	busdata h = Read8(addr);
	if (__predict_false(h.IsBusErr())) {
		return h;
	}
	busdata l = Read8(addr + 1);
	if (__predict_false(l.IsBusErr())) {
		return l;
	}
	// XXX Retry ビットは立たないはず

	busdata data = (h.Data() << 8) | l.Data();
	data.SetWait(h.GetWait() + l.GetWait());
	return data;
}

busdata
IODevice::Read32(uint32 addr)
{
	busdata h = Read16(addr);
	if (__predict_false(h.IsBusErr())) {
		return h;
	}
	busdata l = Read16(addr + 2);
	if (__predict_false(l.IsBusErr())) {
		return l;
	}

	// XXX Retry ビットは立たないはず

	busdata data = (h.Data() << 16) | l.Data();
	data.SetWait(h.GetWait() + l.GetWait());
	return data;
}

busdata
IODevice::Write8(uint32 addr, uint32 data)
{
	return 0;
}

busdata
IODevice::Write16(uint32 addr, uint32 data)
{
	busdata h = Write8(addr, data >> 8);
	if (__predict_false(h.IsBusErr())) {
		return h;
	}
	busdata l = Write8(addr + 1, data & 0xff);
	if (__predict_false(l.IsBusErr())) {
		return h;
	}

	// XXX Retry ビットは立たないはず

	busdata r = 0;
	r.SetWait(h.GetWait() + l.GetWait());
	return r;
}

busdata
IODevice::Write32(uint32 addr, uint32 data)
{
	busdata h = Write16(addr, data >> 16);
	if (__predict_false(h.IsBusErr())) {
		return h;
	}
	busdata l = Write16(addr + 2, data & 0xffff);
	if (__predict_false(l.IsBusErr())) {
		return h;
	}

	// XXX Retry ビットは立たないはず

	busdata r = 0;
	r.SetWait(h.GetWait() + l.GetWait());
	return r;
}

busdata
IODevice::ReadBurst16(uint32 addr, uint32 *dst)
{
	return busdata::BusErr;
}

busdata
IODevice::WriteBurst16(uint32 addr, const uint32 *src)
{
	return busdata::BusErr;
}

busdata
IODevice::Peek8(uint32 addr)
{
	return 0xff;
}

bool
IODevice::Poke8(uint32 addr, uint32 data)
{
	return false;
}

bool
IODevice::Poke(uint32 addr, uint32 data)
{
	return false;
}


//
// バスエラー
//

// コンストラクタ
BusErrDevice::BusErrDevice()
	: inherited(OBJ_BUSERR)
{
	ClearAlias();
	AddAlias("BusErr");
}

// デストラクタ
BusErrDevice::~BusErrDevice()
{
}

busdata
BusErrDevice::Read8(uint32 addr)
{
	putlog(1, "$%08x.b", addr);
	return busdata::BusErr;
}

busdata
BusErrDevice::Read16(uint32 addr)
{
	putlog(1, "$%08x.w", addr);
	return busdata::BusErr;
}

busdata
BusErrDevice::Read32(uint32 addr)
{
	putlog(1, "$%08x.l", addr);
	return busdata::BusErr;
}

busdata
BusErrDevice::Write8(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.b <- %02X", addr, data);
	return busdata::BusErr;
}

busdata
BusErrDevice::Write16(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.w <- %04X", addr, data);
	return busdata::BusErr;
}

busdata
BusErrDevice::Write32(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.l <- %08X", addr, data);
	return busdata::BusErr;
}

busdata
BusErrDevice::Peek8(uint32 addr)
{
	return busdata::BusErr;
}


//
// 何もしないデバイス
//

// コンストラクタ
NopIODevice::NopIODevice()
	: inherited(OBJ_NOPIO)
{
	ClearAlias();
	AddAlias("NopIO");
}

// デストラクタ
NopIODevice::~NopIODevice()
{
}

busdata
NopIODevice::Read8(uint32 addr)
{
	putlog(1, "$%08x.b", addr);
	return 0xff;
}

busdata
NopIODevice::Read16(uint32 addr)
{
	putlog(1, "$%08x.w", addr);
	return 0xffff;
}

busdata
NopIODevice::Read32(uint32 addr)
{
	putlog(1, "$%08x.l", addr);
	return 0xffffffff;
}

busdata
NopIODevice::Write8(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.b <- $%02x", addr, data);
	return 0;
}

busdata
NopIODevice::Write16(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.w <- $%04x", addr, data);
	return 0;
}

busdata
NopIODevice::Write32(uint32 addr, uint32 data)
{
	putlog(1, "$%08x.l <- $%08x", addr, data);
	return 0;
}

busdata
NopIODevice::Peek8(uint32 addr)
{
	return 0xff;
}
